import ipaddress
import itertools
import json
import logging
import re

from flask import Blueprint, Response, current_app

from inventory_provider.routes import common
from inventory_provider.db import opsdb, db

routes = Blueprint("inventory-data-classifier-support-routes", __name__)

logger = logging.getLogger(__name__)


def _LOCATION(equipment, name, abbreviation):
    return {
        'equipment': equipment,
        'name': name,
        'abbreviation': abbreviation
    }


def _remove_duplicates_from_list(all):
    """
    removes duplicates from the input list
    the list items must be encodable as json
    :param l:
    :return: a new list with unique elements
    """
    tmp_dict = dict([(json.dumps(item, sort_keys=True), item) for item in all])
    return list(tmp_dict.values())


def _locations_from_router(router_name):
    r = common.get_current_redis()
    result = r.get(f'opsdb:location:{router_name}')
    if not result:
        logger.error(f'error looking up location for {router_name}')
        return []

    result = json.loads(result.decode('utf-8'))
    if not result:
        logger.error(f'sanity failure: empty list for location {router_name}')
        return []

    return [{
        'a': _LOCATION(
            equipment=result[0]['equipment-name'],
            name=result[0]['pop']['name'],
            abbreviation=result[0]['pop']['abbreviation'])
    }]


def _location_from_service_dict(s):
    location = {
        'a': _LOCATION(
            equipment=s['equipment'],
            name=s['pop_name'],
            abbreviation=s['pop_abbreviation'])
    }

    if all(s[n] for n in (
            'other_end_equipment',
            'other_end_pop_name',
            'other_end_pop_abbreviation')):
        location['b'] = _LOCATION(
            equipment=s['other_end_equipment'],
            name=s['other_end_pop_name'],
            abbreviation=s['other_end_pop_abbreviation'])

    return location


class ClassifierRequestError(Exception):
    status_code = 500

    def __init__(self):
        super().__init__()
        self.message = "Unclassified Internal Error"


class ClassifierProcessingError(ClassifierRequestError):
    status_code = 422

    def __init__(self, message, status_code=None):
        super().__init__()
        self.message = str(message)
        if status_code is not None:
            self.status_code = status_code


@routes.errorhandler(ClassifierRequestError)
def handle_request_error(error):
    return Response(
        response=error.message,
        status=error.status_code)


@routes.after_request
def after_request(resp):
    return common.after_request(resp)


def related_interfaces(hostname, interface):
    r = common.get_current_redis()
    prefix = f'netconf-interfaces:{hostname}:'
    for k in r.keys(f'{prefix}{interface}.*'):
        k = k.decode('utf-8')
        assert k.startswith(prefix)  # sanity
        assert len(k) > len(prefix)  # sanity (contains at least an interface)
        yield k[len(prefix):]


def get_top_level_services(circuit_id, r):
    tls = []
    results = r.get("opsdb:services:children:{}".format(circuit_id))
    if results:
        results = json.loads(results.decode('utf-8'))

        for c in results:
            temp_parents = \
                get_top_level_services(c['parent_circuit_id'], r)
            if not temp_parents:
                tls.append(
                    {'name': c['parent_circuit'],
                     'status': c['parent_circuit_status'],
                     'circuit_type': c['parent_circuit_type'].lower(),
                     'project': c['parent_project']
                     })
            tls.extend(temp_parents)
    return tls


def _link_interface_info(r, hostname, interface):
    """
    Generates the 'interface' field for
    the juniper-link-info response payload.

    only called from get_juniper_link_info

    :param r: redis connection
    :param hostname: router hostname
    :param interface: interface name
    :return: payload dict
    """
    ifc_info = r.get(f'netconf-interfaces:{hostname}:{interface}')
    if not ifc_info:
        # warning: this should match the structure returned by
        #          juniper:list_interfaces:_ifc_info
        ifc_info = {
            'name': interface,
            'description': '',
            'bundle': [],
            'ipv4': [],
            'ipv6': []
        }
    else:
        ifc_info = json.loads(ifc_info.decode('utf-8'))

    bundle_members = r.get(
        f'netconf-interface-bundles:{hostname}:{interface}')
    if bundle_members:
        ifc_info['bundle_members'] \
            = json.loads(bundle_members.decode('utf-8'))
    else:
        ifc_info['bundle_members'] = []

    snmp_info = r.get(
        f'snmp-interfaces-single:{hostname}:{interface}')
    if snmp_info:
        snmp_info = json.loads(snmp_info.decode('utf-8'))
        ifc_info['snmp'] = {
            'community': snmp_info['community'],
            'index': snmp_info['index']
        }

    return ifc_info


@routes.route("/juniper-link-info/<source_equipment>/<path:interface>",
              methods=['GET', 'POST'])
@common.require_accepts_json
def get_juniper_link_info(source_equipment, interface):
    """
    Handler for /classifier/juniper-link-info that
    returns metadata about an IP interface.

    The response will be formatted according to the following schema:

    .. asjson::
       inventory_provider.routes.classifier_schema.JUNIPER_LINK_RESPONSE_SCHEMA

    :param source_equipment: router hostname
    :param interface: link interface name
    :return:
    """

    r = common.get_current_redis()

    cache_key = 'classifier-cache:juniper:%s:%s' % (
        source_equipment, interface)
    result = r.get(cache_key)

    if result:
        result = result.decode('utf-8')
    else:

        result = {
            'locations': []
        }

        top_level_services = []

        services = r.get(
            'opsdb:interface_services:%s:%s' % (source_equipment, interface))
        if services:
            result['services'] = json.loads(services.decode('utf=8'))
            for s in result['services']:
                top_level_services.extend(get_top_level_services(s['id'], r))
            result['locations'] += [
                _location_from_service_dict(s) for s in result['services']]

        result['interface'] = _link_interface_info(
            r, source_equipment, interface)

        def _related_services():
            for related in related_interfaces(source_equipment, interface):
                rs = r.get('opsdb:interface_services:%s:%s'
                           % (source_equipment, related))
                if rs:
                    for s in json.loads(rs.decode('utf-8')):
                        top_level_services.extend(
                            get_top_level_services(s['id'], r))
                        yield {
                            'name': s['name'],
                            'status': s['status'],
                            'circuit_type': s['circuit_type'],
                            'project': s['project']
                        }

        related_services = list(_related_services())
        if related_services:
            top_level_services.extend(related_services)
        if top_level_services:
            result['related-services'] = top_level_services

        if not result['locations']:
            result['locations'] = _locations_from_router(source_equipment)

        result['locations'] = _remove_duplicates_from_list(result['locations'])
        result = json.dumps(result)
        # cache this data for the next call
        r.set(cache_key, result.encode('utf-8'))

    return Response(result, mimetype='application/json')


def _asn_group_info(redis, address):
    """
    :param redis: a redis db connection
    :param address: the remote peer address
    :return:
    """

    try:
        address = ipaddress.ip_address(address).exploded
    except ValueError:
        raise ClassifierProcessingError(
            f'unable to parse {address} as an ip address')

    all_peerings = redis.get(f'juniper-peerings:remote:{address}')
    if not all_peerings:
        return None

    all_peerings = json.loads(all_peerings.decode('utf-8'))
    all_asn = {p['remote-asn'] for p in all_peerings if 'remote-asn' in p}
    if not all_asn:
        return None

    peer_asn = all_asn.pop()
    if all_asn:
        logger.error(
            f'found multiple asn''s for {address}, '
            f'using {peer_asn} and ignoring {all_asn}')

    peerings_this_asn = redis.get(f'juniper-peerings:peer-asn:{peer_asn}')
    if not peerings_this_asn:
        logger.error(
            f'internal data corruption, no peerings found for asn {peer_asn}')
        return None

    peerings_this_asn = json.loads(peerings_this_asn.decode('utf-8'))
    return {
        'asn': peer_asn,
        'peers': [
            {
                'router': p['hostname'],
                'address': p['address'],
                # for future use (in case isolation
                # should be computed per peering type)
                'group': p['group']
            }
            for p in peerings_this_asn
        ]
    }


def _vpn_rr_peering_info(redis, address):
    """
    :param redis: a redis db connection
    :param address: the remote peer address
    :return:
    """
    def _is_rr(peering_info):
        if peering_info.get('logical-system', '') != 'VRR':
            return False
        group = peering_info.get('group', '')
        if group not in ('VPN-RR', 'VPN-RR-INTERNAL'):
            return False
        if 'description' not in peering_info:
            logger.error('internal data error, looks like vpn rr peering'
                         f'but description is missing: {peering_info}')
            return False
        return True

    try:
        address = ipaddress.ip_address(address).exploded
    except ValueError:
        raise ClassifierProcessingError(
            f'unable to parse {address} as an ip address')

    all_peerings = redis.get(f'juniper-peerings:remote:{address}')
    if not all_peerings:
        return None

    all_peerings = json.loads(all_peerings.decode('utf-8'))
    rr_peerings = list(filter(_is_rr, all_peerings))
    if not rr_peerings:
        return None

    if len(rr_peerings) > 1:
        logger.warning(
            f'using the first of multiple vpn rr peer matches: {rr_peerings}')

    return_value = {
        'name': address,
        'description': rr_peerings[0]['description'],
        'router': rr_peerings[0]['hostname']
    }

    if 'remote-asn' in rr_peerings[0]:
        return_value['peer-as'] = rr_peerings[0]['remote-asn']

    return return_value


def _ix_peering_info(redis, address):
    """
    :param redis: a redis db connection
    :param address: the remote peer address
    :return:
    """

    def _is_ix(peering_info):
        if peering_info.get('instance', '') != 'IAS':
            return False
        if not peering_info.get('group', '').startswith('GEANT-IX'):
            return False

        expected_keys = ('description', 'local-asn', 'remote-asn')
        if any(peering_info.get(x, None) is None for x in expected_keys):
            logger.error('internal data error, looks like ix peering but'
                         f'some expected keys are missing: {peering_info}')
            return False
        return True

    try:
        address = ipaddress.ip_address(address).exploded
    except ValueError:
        raise ClassifierProcessingError(
            f'unable to parse {address} as an ip address')

    all_peerings = redis.get(f'juniper-peerings:remote:{address}')
    if not all_peerings:
        return None

    all_peerings = json.loads(all_peerings.decode('utf-8'))
    ix_peerings = list(filter(_is_ix, all_peerings))
    if not ix_peerings:
        return None

    if len(ix_peerings) > 1:
        logger.warning(
            f'using the first of multiple ix peer matches: {ix_peerings}')

    peer_info = {
        'name': address,
        'description': ix_peerings[0]['description'],
        'as': {
            'local': ix_peerings[0]['local-asn'],
            'peer': ix_peerings[0]['remote-asn']
        },
        'router': ix_peerings[0]['hostname']
    }

    return_value = {
        'peer': peer_info,
        'group': [],
        'router': []
    }

    # load the other peers in the same group
    # regex needed??? (e.g. tabs???)
    peering_group_name = peer_info['description'].split(' ')[0]
    peering_group = redis.get(
        f'juniper-peerings:ix-groups:{peering_group_name}')
    if peering_group:
        peering_group = peering_group.decode('utf-8')
        return_value['group'] = sorted(json.loads(peering_group))

    # load the other ix peers from the same router
    router_peerings = redis.get(
        f'juniper-peerings:hosts:{peer_info["router"]}')
    router_peerings = json.loads(router_peerings.decode('utf-8'))
    router_ix_peers = list(filter(_is_ix, router_peerings))
    if router_ix_peers:
        addresses = {p['address'] for p in router_ix_peers}
        return_value['router'] = sorted(list(addresses))

    return return_value


def find_interfaces(address):
    """
    TODO: this is probably the least efficient way of doing this
          (if it's a problem, pre-compute these lists)

    :param address: an ipaddress object
    :return:
    """
    r = common.get_current_redis()
    for k in r.keys('subnets:*'):
        k = k.decode('utf-8')
        m = re.match(r'^subnets:(.*)$', k)
        assert m, 'sanity failure: redis returned an invalid key name'
        interface = ipaddress.ip_interface(m.group(1))
        if address in interface.network:
            info = r.get(k).decode('utf-8')
            info = json.loads(info)
            for ifc in info:
                yield ifc


def find_interfaces_and_services(address_str):
    """

    :param address_str: an ipaddress object
    :return:
    """

    try:
        address = ipaddress.ip_address(address_str)
    except ValueError:
        raise ClassifierProcessingError(
            'unable to parse %r as an ip address' % address_str)

    r = common.get_current_redis()
    for interface in find_interfaces(address):

        services = r.get(
            'opsdb:interface_services:%s:%s' % (
                interface['router'],
                interface['interface name']))
        if not services:
            services = []
        else:
            services = json.loads(services.decode('utf=8'))

        yield {
            'interface': interface,
            'services': services
        }


@routes.route("/peer-info/<address>", methods=['GET', 'POST'])
@common.require_accepts_json
def get_bgp_peer_info(address):
    """
    Handler for /classifier/peer-info that returns bgp peering metadata.

    The response will be formatted according to the following schema:

    .. asjson::
       inventory_provider.routes.classifier_schema.PEER_INFO_RESPONSE_SCHEMA

    :param address: string representation of a bgp peer address
    :return:
    """
    # canonicalize the input address first ...
    try:
        obj = ipaddress.ip_address(address)
        address = obj.exploded
    except ValueError:
        raise ClassifierProcessingError(
            f'unable to parse {address} as an ip address')

    r = common.get_current_redis()

    cache_key = f'classifier-cache:peer:{address}'

    result = r.get(cache_key)
    if result:
        result = result.decode('utf-8')
    else:

        result = {
            'locations': []
        }

        ix_peering_info = _ix_peering_info(r, address)
        if ix_peering_info:
            result['ix-public-peer-info'] = ix_peering_info
            result['locations'] += _locations_from_router(
                ix_peering_info['peer']['router'])

        vpn_rr_peering_info = _vpn_rr_peering_info(r, address)
        if vpn_rr_peering_info:
            result['vpn-rr-peer-info'] = vpn_rr_peering_info
            result['locations'] += _locations_from_router(
                vpn_rr_peering_info['router'])

        asn_group_info = _asn_group_info(r, address)
        if asn_group_info:
            result['asn'] = asn_group_info

        interfaces = list(find_interfaces_and_services(address))
        if interfaces:
            result['interfaces'] = interfaces
            for i in interfaces:
                result['locations'] += [
                    _location_from_service_dict(s) for s in i['services']]

        snmp_info = r.get(
            f'snmp-peerings:remote:{address}')
        if snmp_info:
            snmp_info = json.loads(snmp_info.decode('utf-8'))
            result['snmp'] = [
                {
                    'hostname': h['hostname'],
                    'community': h['community'],
                    'oid': h['oid']
                } for h in snmp_info]

        result['locations'] = _remove_duplicates_from_list(result['locations'])
        result = json.dumps(result)
        # cache this data for the next call
        r.set(cache_key, result.encode('utf-8'))

    return Response(result, mimetype='application/json')


@routes.route("/mtc-interface-info/<node>/<interface>", methods=['GET', 'POST'])
@common.require_accepts_json
def get_mtc_interface_info(node, interface):
    """
    Handler for /classifier/mtc-interface-info that
    returns metadata for as MTC port.

    The response will be formatted according to the following schema:

    .. asjson::
       inventory_provider.routes.classifier_schema.MTC_INTERFACE_INFO_RESPONSE_SCHEMA

    :param node:
    :param element:
    :return: element info
    """
    dummy_response = {
        "contacts": [
            "NOC-RENATER@NOC.RENATER.FR",
            "NOC@GEANT.ORG",
            "NOC@RESTENA.LU",
            "OC@GEANT.ORG",
            "SUPPORT@OC.GEANT.NET"
        ],
        "locations": [
            {
                "a": {
                    "abbreviation": "LON",
                    "equipment": node,
                    "name": "LONDON"
                }
            }
        ],
        "related-services": [
            {
                "circuit_type": "service",
                "name": "RENATER_AP2_CLS",
                "project": "RENATER",
                "status": "operational"
            }
        ],
        "services": [
            {
                "card_id": "",
                "circuit_type": "circuit",
                "equipment": node,
                "id": 669120,
                "logical_unit": "",
                "name": "AMS-LON_EEX_ESNET_1424",
                "other_end_equipment": "",
                "other_end_pop_abbreviation": "",
                "other_end_pop_name": "",
                "other_end_port": "",
                "pop_abbreviation": "LON",
                "pop_name": "LONDON",
                "port": interface,
                "project": "RENATER",
                "service_type": "ETHERNET",
                "status": "operational"
            }
        ]
    }
    return Response(json.dumps(dummy_response), mimetype='application/json')


@routes.route("/infinera-lambda-info/"
              "<source_equipment>/<interface>/<circuit_id>",
              methods=['GET', 'POST'])
@common.require_accepts_json
def get_infinera_lambda_info(source_equipment, interface, circuit_id):
    """
    Handler for /classifier/infinera-lambda-info that
    returns metadata for a DTNX port.

    The response will be formatted according to the following schema:

    .. asjson::
       inventory_provider.routes.classifier_schema.INFINERA_LAMBDA_INFO_RESPONSE_SCHEMA

    :param source_equipment: DTNX name
    :param address: interface/port name
    :param circuit_id: infinera circuit id
    :return:
    """

    interface = interface.replace('-T', '-')
    cache_key = 'classifier-cache:infinera:%s:%s' % (
        source_equipment, interface)

    r = common.get_current_redis()
    result = r.get(cache_key)

    if result:
        result = result.decode('utf-8')
    else:
        result = {
            'locations': []
        }
        top_level_services = []

        services = r.get(
            'opsdb:interface_services:%s:%s' % (source_equipment, interface))
        if services:
            result['services'] = json.loads(services.decode('utf=8'))
            for s in result['services']:
                tls = get_top_level_services(s['id'], r)
                if tls:
                    top_level_services.extend(tls)
            if top_level_services:
                result['related-services'] = top_level_services
            result['locations'] += [
                _location_from_service_dict(s) for s in result['services']]

        gl = r.get('opsdb:geant_lambdas:%s' % circuit_id.lower())
        if gl:
            t_gl = json.loads(gl.decode('utf=8'))
            t_gl['status'] = t_gl['status'].lower()
            result['geant-lambda'] = t_gl

        if not result:
            return Response(
                response="no available info for {} {}".format(
                    source_equipment, interface),
                status=404,
                mimetype="text/html")

        result = json.dumps(result)
        # cache this data for the next call
        r.set(cache_key, result.encode('utf-8'))

    return Response(result, mimetype='application/json')


@routes.route("/infinera-fiberlink-info/<ne_name_str>/<object_name_str>",
              methods=['GET', 'POST'])
@common.require_accepts_json
def get_fiberlink_trap_metadata(ne_name_str, object_name_str):
    """
    Handler for /classifier/infinera-fiberlink-info that
    returns metadata for a particular opitical path segment.

    TODO: no schema is declared, and there are no validation tests

    :param ne_name_str: OLA or DTNX equipment name
    :param object_name_str: path name
    :return:
    """

    objects = object_name_str.split('_')
    shelves = [x.split('-')[0] for x in objects]
    p = r'([a-zA-Z\d]+?-(OLA|DTNX)\d+(-\d)?)'
    matches = re.findall(p, ne_name_str)
    if len(matches) != 2 or len(shelves) != 2:
        raise ClassifierProcessingError(
            f'unable to parse {ne_name_str} {object_name_str } '
            'into two elements')

    r = common.get_current_redis()

    # double check that we only need to check the two nodes and not the objects
    cache_key = f'classifier-cache:fiberlink:{ne_name_str}:{object_name_str}'
    result = r.get(cache_key)

    if result:
        result = result.decode('utf-8')
    else:
        equipment_a = matches[0][0]
        equipment_b = matches[1][0]
        nes_a = f'{equipment_a}-{shelves[0]}'
        nes_b = f'{equipment_b}-{shelves[1]}'
        result = []
        df_a = r.get(f'opsdb:ne_fibre_spans:{nes_a}')
        df_b = r.get(f'opsdb:ne_fibre_spans:{nes_b}')
        if df_a and df_b:
            a = json.loads(df_a.decode('utf-8'))
            b = json.loads(df_b.decode('utf-8'))

            matches = [x for x in itertools.product(a, b) if
                       x[0]['df_route_id'] == x[1]['df_route_id']]
            if matches:
                match = matches[0]
                result = {
                    'ends': {
                        'a': {
                            'pop': match[0]['pop'],
                            'pop_abbreviation': match[0]['pop_abbreviation'],
                            'equipment': equipment_a
                        },
                        'b': {
                            'pop': match[1]['pop'],
                            'pop_abbreviation': match[1]['pop_abbreviation'],
                            'equipment': equipment_b
                        },
                    },
                    'df_route': {
                        'id': match[0]['df_route_id'],
                        'name': match[0]['df_route'],
                        'status': match[0]['df_status'],
                    },
                    'related-services':
                        get_top_level_services(match[0]['df_route_id'], r)
                }
                result = json.dumps(result)
                r.set(cache_key, result)
        if not result:
            return Response(
                response="no available info for "
                         f"{ne_name_str} {object_name_str}",
                status=404,
                mimetype="text/html")

    return Response(result, mimetype='application/json')


@routes.route('/coriant-info/<equipment_name>/<path:entity_string>',
              methods=['GET', 'POST'])
@common.require_accepts_json
def get_coriant_info(equipment_name, entity_string):
    """
    Handler for /classifier/coriant-info that
    returns metadata for a coriant path.

    The response will be formatted according to the following schema:

    .. asjson::
       inventory_provider.routes.classifier_schema.CORIANT_INFO_RESPONSE_SCHEMA

    :param source_equipment: grv hostname
    :param entity_string: path name
    :return:
    """

    r = common.get_current_redis()

    cache_key = 'classifier-cache:coriant:%s:%s' % (
        equipment_name, entity_string)
    result = r.get(cache_key)

    if result:
        result = result.decode('utf-8')
    else:

        m = re.match(r'^(\d+\-\d+)\.(\d+)', entity_string)
        if not m:
            logger.error(
                'invalid coriant entity string format: %r' % entity_string)
            return Response(
                response="no available info for '{}' '{}'".format(
                    equipment_name, entity_string),
                status=404,
                mimetype="text/html")
        result = {
            'equipment name': equipment_name,
            'card id': m.group(1),
            'port number': m.group(2),
            'locations': []
        }

        config = current_app.config['INVENTORY_PROVIDER_CONFIG']

        with db.connection(config['ops-db']) as cx:
            path = opsdb.lookup_coriant_path(
                cx, equipment_name, result['card id'], result['port number'])

        if path:
            result['path'] = path
            result['locations'] += [{
                'a': _LOCATION(
                    equipment=path['a']['equipment name'],
                    name=path['a']['pop']['name'],
                    abbreviation=path['a']['pop']['abbreviation']),
                'b': _LOCATION(
                    equipment=path['b']['equipment name'],
                    name=path['b']['pop']['name'],
                    abbreviation=path['b']['pop']['abbreviation']),
            }]

            top_level_services = get_top_level_services(path['id'], r)
            if top_level_services:
                result['related-services'] = top_level_services
        else:
            logger.error('no path found for {}:{}'.format(
                equipment_name, entity_string))

        # cache this data for the next call
        result = json.dumps(result).encode('utf-8')
        r.set(cache_key, result)

    return Response(result, mimetype='application/json')
